"""Raresim: A Python interface for scalable rare-variant simulations.

This package provides tools for simulating and analyzing rare genetic variants
in population genetics studies.
"""

__version__ = "3.0.0"
__all__ = ['__version__']